import unittest.mock


def _test_module_init(module, main_name="main"):
    with unittest.mock.patch.object(
        module, main_name, return_value=0
    ), unittest.mock.patch.object(
        module, "__name__", "__main__"
    ), unittest.mock.patch.object(
        module.sys, "exit"
    ) as exit:
        module.module_init()
        return exit.call_args[0][0] == 0



import time
import unittest


class TestPython(unittest.TestCase):
    def test_monotonic(self):
        val1 = time.monotonic()
        val2 = time.monotonic()
        self.assertGreaterEqual(val2, val1)



import platform
import unittest

from twoping import packets, utils


class TestUtils(unittest.TestCase):
    def test_twoping_checksum_hello(self):
        data = b"Hello World"
        self.assertEqual(utils.twoping_checksum(data), 0xAE31)

    def test_twoping_checksum_junk(self):
        data = (
            b"\x45\xc6\xca\x92\x10\x0e\xb1\xcf\x98\x88\x0b\x42\xc0\xf8\x58\xac\xd9\x81\x76\xc0\x45\x8c\x6f"
            b"\x04\x9c\x0e\x93\xb3\x49\x3d\x38\x6d\xb7\xd5\x86\xa7\x66\x2c\x32\x15\xd9\x7f\xad\x3e\x1a\xe0"
            b"\x39\x89\x4a\xdd\x0b\x28\xa0\x07\x61\x47\x54\xc6\x50\x04\x2d\x28\x4e\x48\x3c\x40\xc7\xdb\x82"
            b"\xfa\x0b\xab\x37\x6e\x23\x5e\x80\xab\xb9\xe5\x05\x14\xb1\xfc\x72\x54\x37\x8a\x66\xf2\x51\xa1"
            b"\x77\xda\xd5\x88\xd5\x38\x43\x31\x30\xe7\x1e\x7a\x2b\xc8\x14\x0d\x4b\x7c\x33\x9a\x9b\xc1\xc1"
            b"\xc8\xc5\xfe\x5a\x48\x44\x1b\x87\x9d\x24\xd9\x22\x7d"
        )
        self.assertEqual(utils.twoping_checksum(data), 0x4A06)

    def test_twoping_checksum_packet(self):
        data = bytearray(
            b"\x32\x50\x8d\x3b\x00\x00\x00\x00\xb0\x06\x00\x3b\x00\x00\x00\x06\x00\x00\x00\x00\xa0\x0a\x00"
            b"\x08\x00\x01\x00\x00\x00\x00\xa0\x01\x00\x08\x00\x01\x00\x00\x00\x00\xa0\x02\x00\x08\x00\x01"
            b"\x00\x00\x00\x00\xb0\x02"
        )
        data[2] = 0
        data[3] = 0
        self.assertEqual(utils.twoping_checksum(data), 0x8D3B)

    def test_twoping_checksum_iter(self):
        i = 65535
        for x in range(256):
            for y in range(256):
                data = bytes([x, y])
                checksum = utils.twoping_checksum(data)
                self.assertEqual(checksum, i)
                i = i - 1
                if i == 0:
                    i = 65535

    def test_fuzz_bytearray(self):
        data = packets.Packet().dump()
        data_fuzzed = bytearray(data)
        utils.fuzz_bytearray(data_fuzzed, 100)
        self.assertNotEqual(data, bytes(data_fuzzed))

    def test_fuzz_bytearray_zero(self):
        data = packets.Packet().dump()
        data_fuzzed = bytearray(data)
        utils.fuzz_bytearray(data_fuzzed, 0)
        self.assertEqual(data, bytes(data_fuzzed))

    def test_fuzz_packet(self):
        data = packets.Packet().dump()
        data_fuzzed = utils.fuzz_packet(data, 100)
        self.assertNotEqual(data, data_fuzzed)

    def test_fuzz_packet_zero(self):
        data = packets.Packet().dump()
        data_fuzzed = utils.fuzz_packet(data, 0)
        self.assertEqual(data, data_fuzzed)

    def test_div0(self):
        self.assertEqual(utils.div0(10, 5), 10 / 5)
        self.assertEqual(utils.div0(5, 0), 0)
        self.assertEqual(utils.div0(0, 0), 0)

    def test_npack(self):
        self.assertEqual(utils.npack(1), b"\x01")
        self.assertEqual(utils.npack(1234), b"\x04\xd2")
        self.assertEqual(utils.npack(123456), b"\x01\xe2\x40")

    def test_npack_minimum(self):
        self.assertEqual(utils.npack(1, 2), b"\x00\x01")
        self.assertEqual(utils.npack(1234, 2), b"\x04\xd2")
        self.assertEqual(utils.npack(123456, 4), b"\x00\x01\xe2\x40")

    def test_nunpack(self):
        self.assertEqual(utils.nunpack(b"\x01"), 1)
        self.assertEqual(utils.nunpack(b"\x04\xd2"), 1234)
        self.assertEqual(utils.nunpack(b"\x00\x00\x04\xd2"), 1234)
        self.assertEqual(utils.nunpack(b"\x01\xe2\x40"), 123456)
        self.assertEqual(utils.nunpack(b"\x00\x01\xe2\x40"), 123456)

    def test_platform_info(self):
        self.assertIn(platform.system(), utils.platform_info())

    def test_stats_time(self):
        self.assertEqual(utils.stats_time(123.45), "2m 3s 450ms")
        self.assertEqual(utils.stats_time(0.45678), "456ms")
        self.assertEqual(utils.stats_time(123456789000), "3914y 288d 30m")



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2021 Ryan Finnie
# SPDX-License-Identifier: GPL-2.0+

import errno
import logging
import math
import selectors
import signal
import socket
import sys
import time

try:
    import dns.resolver as dns_resolver
except ImportError as e:
    dns_resolver = e

try:
    import netifaces
except ImportError as e:
    netifaces = e

try:
    import systemd.daemon as systemd_daemon
except ImportError as e:
    systemd_daemon = e

from . import __version__, packets
from .args import parse_args
from .utils import (
    _,
    _pl,
    fuzz_packet,
    div0,
    platform_info,
    random,
    random_is_systemrandom,
    stats_time,
)


version_string = "2ping {} - {}".format(__version__, platform_info())
clock = time.perf_counter


class SocketClass:
    def __init__(self, sock):
        self.sock = sock
        self.address = sock.getsockname()
        if not self.address:
            self.address = None

        # Functional tags applied to the SocketClass
        self.is_client = False
        self.is_listener = False
        self.is_loopback = False
        self.is_inet = False
        self.is_systemd = False

        # Used during client mode for the host tuple to send UDP packets to.
        self.client_host = None
        # (family, type, proto, canonname, sockaddr) of the bound socket
        self.bind_addrinfo = None
        # Whether the sock has been closed
        self.closed = False

        # Dict of PeerState instances, indexed by peer tuple
        self.peer_states = {}

        # Statistics
        self.pings_transmitted = 0
        self.pings_received = 0
        self.packets_transmitted = 0
        self.packets_received = 0
        self.lost_outbound = 0
        self.lost_inbound = 0
        self.errors_received = 0
        self.rtt_total = 0
        self.rtt_total_sq = 0
        self.rtt_count = 0
        self.rtt_min = 0
        self.rtt_max = 0
        self.rtt_ewma = 0

        self.next_send = 0

        self.shutdown_time = 0
        self.session = bytes([random.randint(0, 255) for x in range(8)])

    def __repr__(self):
        attrs = ["fd {}".format(self.fileno())]
        if self.bind_addrinfo:
            attrs.append("bind {}".format(self.bind_addrinfo))
        if self.client_host:
            attrs.append("client {}".format(self.client_host))
        for tag in ("client", "listener", "inet", "loopback", "systemd"):
            if getattr(self, "is_{}".format(tag)):
                attrs.append(tag)
        if self.closed:
            attrs.append("closed")
        return "<SocketClass: {}>".format(", ".join(attrs))

    def fileno(self):
        return self.sock.fileno()


class PeerState:
    def __init__(self, peer_tuple, sock_class):
        self.peer_tuple = peer_tuple
        self.sock_class = sock_class

        # In-flight outbound messages.  Added in the following conditions:
        #   * Outbound packet with OpcodeReplyRequested sent.
        # Removed in following conditions:
        #   * Inbound packet with OpcodeInReplyTo set to it.
        #   * Inbound packet with it in OpcodeInvestigationSeen or OpcodeInvestigationUnseen.
        #   * Cleanup after 10 minutes.
        # If it remains for more than <10> seconds, it it sent as part of
        # OpcodeInvestigate with the next outbound packet with OpcodeReplyRequested set.
        self.sent_messages = {}
        # Seen inbound messages.  Added in the following conditions:
        #   * Inbound packet with OpcodeReplyRequested set.
        # Referenced in the following conditions:
        #   * Inbound packet with it in OpcodeInvestigate.
        # Removed in the following conditions:
        #   * Inbound packet with it in OpcodeCourtesyExpiration.
        #   * Cleanup after 10 minutes.
        self.seen_messages = {}
        # Courtesy messages waiting to be sent.  Added in the following conditions:
        #   * Inbound packet with OpcodeInReplyTo set.
        # Removed in the following conditions:
        #   * Outbound packet where there is room to send it as part of OpcodeCourtesyExpiration.
        #   * Cleanup after 2 minutes.
        self.courtesy_messages = {}
        # Current position of a peer tuple's incrementing ping integer.
        self.ping_position = 0
        # Current encrypted session ID
        self.encrypted_session_id = None
        # Seen encrypted session IVs
        self.encrypted_session_ivs = {}
        # Last time a peer was sent to or received from
        self.last_seen = clock()


class TwoPing:
    def __init__(self, args):
        now = clock()
        self.args = args
        self.time_start = now
        self.fake_time_epoch = random.random() * (2 ** 32)
        self.fake_time_generation = random.randint(0, 65535)
        self.isatty = sys.stdout.isatty()

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)

        self.sock_classes = []
        self.poller = selectors.DefaultSelector()

        # Scheduled events
        self.next_cleanup = now + 60.0
        self.next_stats = 0
        if self.args.stats:
            self.next_stats = now + self.args.stats

        self.old_age_interval = 60.0
        # On Windows, KeyboardInterrupt during select() will not be trapped
        # until a socket event or timeout, so we should set the timeout to
        # a short value.
        if sys.platform.startswith(("win32", "cygwin")):
            self.old_age_interval = 1.0

        # Test for IPv6 functionality.
        self.has_ipv6 = True
        if not socket.has_ipv6:
            self.has_ipv6 = False
        else:
            # BSD jails seem to have has_ipv6 = True, but will throw
            # "Protocol not supported" on bind.  Test for this.
            try:
                socket.socket(socket.AF_INET6, socket.SOCK_DGRAM).close()
            except socket.error as e:
                if e.errno == errno.EPROTONOSUPPORT:
                    self.has_ipv6 = False
                else:
                    raise

        self.is_reload = False

    def print_out(self, *args, **kwargs):
        print(*args, **kwargs)

    def tty_out(self, *args, **kwargs):
        if not self.isatty:
            return
        self.print_out(*args, **kwargs)

    @property
    def sock_classes_open(self):
        return [sock_class for sock_class in self.sock_classes if not sock_class.closed]

    @property
    def sock_classes_active(self):
        now = clock()
        return [
            sock_class
            for sock_class in self.sock_classes_open
            if not sock_class.shutdown_time
            or (sock_class.shutdown_time and sock_class.shutdown_time > now)
        ]

    def handle_socket_error(self, e, sock_class, peer_address=None):
        sock = sock_class.sock
        # Errors from the last send() can be trapped via IP_RECVERR (Linux only).
        sock_class.errors_received += 1
        error_string = str(e)
        try:
            MSG_ERRQUEUE = 8192
            (error_data, error_address) = sock.recvfrom(16384, MSG_ERRQUEUE)
            print_address = error_address[0]
        except socket.error:
            if peer_address:
                print_address = peer_address[0]
            else:
                print_address = str(sock_class)
        if self.args.quiet:
            pass
        elif print_address:
            self.logger.error("{}: {}".format(print_address, error_string))
        else:
            self.logger.error(error_string)

    def process_incoming_packet(self, sock_class):
        sock = sock_class.sock
        recvfrom = self.sock_recvfrom(sock_class)
        time_begin = clock()
        if recvfrom is None:
            # Error handled, simulated packet loss, etc
            return
        data, peer_address = recvfrom
        if peer_address:
            print_address = peer_address[0]
        else:
            print_address = str(sock_class)
        socket_address = sock_class.address
        self.logger.debug("Socket: {}, peer: {}".format(sock_class, peer_address))

        # Per-packet options.
        sock_class.packets_received += 1
        calculated_rtt = None

        peer_tuple = (socket_address, peer_address, sock.type)

        # Preload state tables if the client has not been seen (or has been cleaned).
        if peer_tuple not in sock_class.peer_states:
            sock_class.peer_states[peer_tuple] = PeerState(peer_tuple, sock_class)
        peer_state = sock_class.peer_states[peer_tuple]
        peer_state.last_seen = time_begin

        # Load/parse the packet.
        packet_in = packets.Packet()
        packet_in.load(data)
        if self.args.verbose:
            self.logger.info("RECV: {}".format(packet_in))

        # Decrypt packet if needed.
        if self.args.encrypt:
            if packets.OpcodeEncrypted.id not in packet_in.opcodes:
                sock_class.errors_received += 1
                self.logger.error(
                    _("Encryption required but not provided by {address}").format(
                        address=print_address
                    )
                )
                return
            if (
                packet_in.opcodes[packets.OpcodeEncrypted.id].method_index
                != self.args.encrypt_method_index
            ):
                sock_class.errors_received += 1
                self.logger.error(
                    _(
                        "Encryption method mismatch from {address} (expected {expected}, got {got})"
                    ).format(
                        address=print_address,
                        expected=self.args.encrypt_method_index,
                        got=packet_in.opcodes[packets.OpcodeEncrypted.id].method_index,
                    )
                )
                return
            data = packet_in.opcodes[packets.OpcodeEncrypted.id].decrypt(
                self.args.encrypt.encode("UTF-8")
            )
            encrypted_packet_in = packet_in
            packet_in = packets.Packet()
            packet_in.load(data)
            if peer_state.encrypted_session_id is None:
                peer_state.encrypted_session_id = encrypted_packet_in.opcodes[
                    packets.OpcodeEncrypted.id
                ].session
            if (
                encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].session
                != peer_state.encrypted_session_id
            ):
                sock_class.errors_received += 1
                self.logger.error(
                    _(
                        "Encryption session mismatch from {address} (expected {expected}, got {got})"
                    ).format(
                        address=print_address,
                        expected=peer_state.encrypted_session_id,
                        got=encrypted_packet_in.opcodes[
                            packets.OpcodeEncrypted.id
                        ].session,
                    )
                )
                return
            if (
                encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].iv
                in peer_state.encrypted_session_ivs
            ):
                sock_class.errors_received += 1
                self.logger.error(
                    _("Repeated IV {iv} from {address}, discarding").format(
                        iv=encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].iv,
                        address=print_address,
                    )
                )
                return
            peer_state.encrypted_session_ivs[
                encrypted_packet_in.opcodes[packets.OpcodeEncrypted.id].iv
            ] = (time_begin,)
            if self.args.verbose:
                self.logger.info("DECR: {}".format(packet_in))

        # Verify HMAC if required.
        if self.args.auth:
            if packets.OpcodeHMAC.id not in packet_in.opcodes:
                sock_class.errors_received += 1
                self.logger.error(
                    _("Auth required but not provided by {address}").format(
                        address=print_address
                    )
                )
                return
            if (
                packet_in.opcodes[packets.OpcodeHMAC.id].digest_index
                != self.args.auth_digest_index
            ):
                sock_class.errors_received += 1
                self.logger.error(
                    _(
                        "Auth digest type mismatch from {address} (expected {expected}, got {got})"
                    ).format(
                        address=print_address,
                        expected=self.args.auth_digest_index,
                        got=packet_in.opcodes[packets.OpcodeHMAC.id].digest_index,
                    )
                )
                return
            (test_begin, test_length) = packet_in.opcode_data_positions[
                packets.OpcodeHMAC.id
            ]
            test_begin += 2
            test_length -= 2
            packet_in.opcodes[packets.OpcodeHMAC.id].key = self.args.auth.encode(
                "UTF-8"
            )
            test_data = bytearray(data)
            test_data[2:4] = b"\x00\x00"
            test_data[test_begin : (test_begin + test_length)] = bytes(test_length)
            test_hash = packet_in.opcodes[packets.OpcodeHMAC.id].hash
            test_hash_calculated = packet_in.calculate_hash(
                packet_in.opcodes[packets.OpcodeHMAC.id], test_data
            )
            if test_hash_calculated != test_hash:
                sock_class.errors_received += 1
                self.logger.error(
                    _(
                        "Auth hash failed from {address} (expected {expected}, got {got})"
                    ).format(
                        address=print_address,
                        expected=test_hash_calculated.hex(),
                        got=test_hash.hex(),
                    )
                )
                return

        # If this is in reply to one of our sent packets, it's a ping reply, so handle it specially.
        if packets.OpcodeInReplyTo.id in packet_in.opcodes:
            replied_message_id = packet_in.opcodes[
                packets.OpcodeInReplyTo.id
            ].message_id
            if replied_message_id in peer_state.sent_messages:
                (sent_time, _unused, ping_position) = peer_state.sent_messages[
                    replied_message_id
                ]
                del peer_state.sent_messages[replied_message_id]

                calculated_rtt = (time_begin - sent_time) * 1000
                if (
                    self.args.subtract_peer_host_latency
                    and packets.OpcodeHostLatency.id in packet_in.opcodes
                ):
                    calculated_rtt -= (
                        packet_in.opcodes[packets.OpcodeHostLatency.id].delay_us
                        / 1000.0
                    )
                # Peer host latency could be wildly innacurate; guard against negative RTT.
                if calculated_rtt < 0:
                    calculated_rtt = 0

                sock_class.pings_received += 1
                self.update_rtts(sock_class, calculated_rtt)
                if self.args.quiet or self.args.flood:
                    pass
                else:
                    if self.args.audible:
                        self.tty_out("\x07", end="", flush=True)
                    if packets.OpcodeRTTEnclosed.id in packet_in.opcodes:
                        self.logger.info(
                            _(
                                (
                                    "{bytes} bytes from {address}: ping_seq={seq} time={ms:0.03f} ms "
                                    "peertime={peerms:0.03f} ms"
                                )
                            ).format(
                                bytes=len(data),
                                address=print_address,
                                seq=ping_position,
                                ms=calculated_rtt,
                                peerms=(
                                    packet_in.opcodes[
                                        packets.OpcodeRTTEnclosed.id
                                    ].rtt_us
                                    / 1000.0
                                ),
                            )
                        )
                    else:
                        self.logger.info(
                            _(
                                "{bytes} bytes from {address}: ping_seq={seq} time={ms:0.03f} ms"
                            ).format(
                                bytes=len(data),
                                address=print_address,
                                seq=ping_position,
                                ms=calculated_rtt,
                            )
                        )
                    if (packets.OpcodeExtended.id in packet_in.opcodes) and (
                        packets.ExtendedNotice.id
                        in packet_in.opcodes[packets.OpcodeExtended.id].segments
                    ):
                        notice = (
                            packet_in.opcodes[packets.OpcodeExtended.id]
                            .segments[packets.ExtendedNotice.id]
                            .text
                        )
                        self.logger.info(
                            "  " + _("Peer notice: {notice}").format(notice=notice)
                        )
            peer_state.courtesy_messages[replied_message_id] = (
                time_begin,
                replied_message_id,
            )

        # Check if any invesitgations results have come back.
        self.check_investigations(peer_state, packet_in)

        # Process courtesy expirations
        if packets.OpcodeCourtesyExpiration.id in packet_in.opcodes:
            for message_id in packet_in.opcodes[
                packets.OpcodeCourtesyExpiration.id
            ].message_ids:
                if message_id in peer_state.seen_messages:
                    del peer_state.seen_messages[message_id]

        # If the peer requested a reply, prepare one.
        if packets.OpcodeReplyRequested.id in packet_in.opcodes:
            # Populate seen_messages.
            peer_state.seen_messages[packet_in.message_id] = (time_begin,)

            # Basic packet configuration.
            packet_out = self.base_packet()
            packet_out.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
            packet_out.opcodes[
                packets.OpcodeInReplyTo.id
            ].message_id = packet_in.message_id

            # If we are matching packet sizes of the peer, adjust the minimum if it falls between min_packet_size
            # and max_packet_size.
            if not self.args.no_match_packet_size:
                data_len = len(data)
                if (data_len <= self.args.max_packet_size) and (
                    data_len >= self.args.min_packet_size
                ):
                    packet_out.min_length = data_len

            # 3-way pings already have the first roundtrip calculated.
            if calculated_rtt is not None:
                packet_out.opcodes[
                    packets.OpcodeRTTEnclosed.id
                ] = packets.OpcodeRTTEnclosed()
                packet_out.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = int(
                    calculated_rtt * 1000
                )

            # Check for any investigations the peer requested.
            if packets.OpcodeInvestigate.id in packet_in.opcodes:
                for message_id in packet_in.opcodes[
                    packets.OpcodeInvestigate.id
                ].message_ids:
                    if message_id in peer_state.seen_messages:
                        if packets.OpcodeInvestigationSeen.id not in packet_out.opcodes:
                            packet_out.opcodes[
                                packets.OpcodeInvestigationSeen.id
                            ] = packets.OpcodeInvestigationSeen()
                        packet_out.opcodes[
                            packets.OpcodeInvestigationSeen.id
                        ].message_ids.append(message_id)
                    else:
                        if (
                            packets.OpcodeInvestigationUnseen.id
                            not in packet_out.opcodes
                        ):
                            packet_out.opcodes[
                                packets.OpcodeInvestigationUnseen.id
                            ] = packets.OpcodeInvestigationUnseen()
                        packet_out.opcodes[
                            packets.OpcodeInvestigationUnseen.id
                        ].message_ids.append(message_id)

            # If the packet_in is ReplyRequested but not InReplyTo, it is a second leg.  Unless 3-way ping was
            # disabled, request a reply.
            if (packets.OpcodeInReplyTo.id not in packet_in.opcodes) and (
                not self.args.no_3way
            ):
                packet_out.opcodes[
                    packets.OpcodeReplyRequested.id
                ] = packets.OpcodeReplyRequested()

            # Send any investigations we would like to know about.
            self.start_investigations(peer_state, packet_out)

            # Any courtesy expirations we have waiting should be sent.
            if len(peer_state.courtesy_messages) > 0:
                packet_out.opcodes[
                    packets.OpcodeCourtesyExpiration.id
                ] = packets.OpcodeCourtesyExpiration()
                for (
                    courtesy_time,
                    courtesy_message_id,
                ) in peer_state.courtesy_messages.values():
                    packet_out.opcodes[
                        packets.OpcodeCourtesyExpiration.id
                    ].message_ids.append(courtesy_message_id)

            # Calculate the host latency as late as possible.
            packet_out.opcodes[
                packets.OpcodeHostLatency.id
            ] = packets.OpcodeHostLatency()
            time_send = clock()
            packet_out.opcodes[packets.OpcodeHostLatency.id].delay_us = int(
                (time_send - time_begin) * 1000000
            )

            # Dump the packet.
            dump_out = packet_out.dump()

            # If enabled, encrypt the packet and wrap it in a stub packet.
            if self.args.encrypt:
                encrypted_packet = packets.Packet()
                encrypted_packet.opcodes[
                    packets.OpcodeEncrypted.id
                ] = packets.OpcodeEncrypted()
                encrypted_packet.opcodes[
                    packets.OpcodeEncrypted.id
                ].method_index = self.args.encrypt_method_index
                encrypted_packet.opcodes[
                    packets.OpcodeEncrypted.id
                ].session = encrypted_packet_in.opcodes[
                    packets.OpcodeEncrypted.id
                ].session
                encrypted_packet.opcodes[packets.OpcodeEncrypted.id].encrypt(
                    dump_out, self.args.encrypt.encode("UTF-8")
                )
                sock_out = encrypted_packet.dump()
            else:
                sock_out = dump_out

            # Send the packet.
            self.sock_sendto(sock_class, sock_out, peer_address)
            sock_class.packets_transmitted += 1

            # If ReplyRequested is set, we care about its arrival.
            if packets.OpcodeReplyRequested.id in packet_out.opcodes:
                sock_class.pings_transmitted += 1
                peer_state.ping_position += 1
                peer_state.sent_messages[packet_out.message_id] = (
                    time_send,
                    packet_out.message_id,
                    peer_state.ping_position,
                )

            # Examine the sent packet.
            packet_out_examine = packets.Packet()
            packet_out_examine.load(dump_out)

            # Any courtesy expirations which had room in the sent packet should be forgotten.
            if packets.OpcodeCourtesyExpiration.id in packet_out_examine.opcodes:
                for courtesy_message_id in packet_out_examine.opcodes[
                    packets.OpcodeCourtesyExpiration.id
                ].message_ids:
                    if courtesy_message_id in peer_state.courtesy_messages:
                        del peer_state.courtesy_messages[courtesy_message_id]

            if self.args.verbose:
                self.logger.info(
                    "SEND{}: {}".format(
                        (" (encrypted)" if self.args.encrypt else ""),
                        packet_out_examine,
                    )
                )

        if sock_class.next_send and (packets.OpcodeInReplyTo.id in packet_in.opcodes):
            self.schedule_next_send(sock_class, reply_received=True)

    def sock_sendto(self, sock_class, data, address=None):
        sock = sock_class.sock
        # Simulate random packet loss.
        if self.args.packet_loss_out and (
            random.random() < (self.args.packet_loss_out / 100.0)
        ):
            return
        # Send the packet.
        try:
            if address:
                sock.sendto(data, address)
            else:
                sock.send(data)
        except socket.error as e:
            self.handle_socket_error(e, sock_class, peer_address=address)

    def sock_recvfrom(self, sock_class):
        sock = sock_class.sock
        try:
            (data, peer_address) = sock.recvfrom(16384)
        except socket.error as e:
            self.handle_socket_error(e, sock_class)
            return

        # Simulate random packet loss.
        if self.args.packet_loss_in and (
            random.random() < (self.args.packet_loss_in / 100.0)
        ):
            return

        # Simulate data corruption
        if self.args.fuzz:
            data = fuzz_packet(data, self.args.fuzz)

        if not peer_address:
            peer_address = None

        return data, peer_address

    def start_investigations(self, peer_state, packet_check):
        if len(peer_state.sent_messages) == 0:
            return
        if packets.OpcodeInvestigate.id in packet_check.opcodes:
            iobj = packet_check.opcodes[packets.OpcodeInvestigate.id]
        else:
            iobj = None
        now = clock()
        for message_id_str in peer_state.sent_messages:
            (sent_time, message_id, _unused) = peer_state.sent_messages[message_id_str]
            if now >= (sent_time + self.args.inquire_wait):
                if iobj is None:
                    iobj = packets.OpcodeInvestigate()
                if message_id not in iobj.message_ids:
                    iobj.message_ids.append(message_id)
        if iobj is not None:
            packet_check.opcodes[packets.OpcodeInvestigate.id] = iobj

    def check_investigations(self, peer_state, packet_check):
        found = {}

        # Inbound/Outbound
        for opcode_id, type_str, type_stat in [
            (packets.OpcodeInvestigationSeen.id, "inbound", "lost_inbound"),
            (packets.OpcodeInvestigationUnseen.id, "outbound", "lost_outbound"),
        ]:
            if opcode_id in packet_check.opcodes:
                for message_id in packet_check.opcodes[opcode_id].message_ids:
                    if message_id not in peer_state.sent_messages:
                        continue
                    (_unused, _unused, ping_seq) = peer_state.sent_messages[message_id]
                    if peer_state.peer_tuple[1]:
                        address = peer_state.peer_tuple[1][0]
                    else:
                        address = peer_state.sock_class
                    found[ping_seq] = (type_str, address)
                    del peer_state.sent_messages[message_id]
                    setattr(
                        peer_state.sock_class,
                        type_stat,
                        getattr(peer_state.sock_class, type_stat) + 1,
                    )

        if self.args.quiet:
            return
        # Print results
        for ping_seq in sorted(found):
            (loss_type, address) = found[ping_seq]
            if loss_type == "inbound":
                loss_message = "Lost inbound packet from {address}: ping_seq={seq}"
            else:
                loss_message = "Lost outbound packet to {address}: ping_seq={seq}"
            self.logger.error(
                _(loss_message).format(
                    loss_type=loss_type, address=address, seq=ping_seq
                )
            )

    def close_socks(self, sock_classes=()):
        if sock_classes == ():
            sock_classes = self.sock_classes_open
        for sock_class in sock_classes:
            self.logger.debug("Closing socket: {}".format(sock_class))
            self.poller.unregister(sock_class)
            sock_class.sock.close()
            sock_class.closed = True
            sock_class.shutdown_time = clock()
            sock_class.next_send = 0

    def gather_systemd_socks(self):
        # Do nothing if systemd.daemon is not available
        if isinstance(systemd_daemon, ImportError):
            return []

        # If we've done this before, don't re-attempt
        systemd_sock_classes = [
            sock_class for sock_class in self.sock_classes if sock_class.is_systemd
        ]
        if systemd_sock_classes:
            return systemd_sock_classes

        # Note that listen_fds() defaults to unset_environment=True
        # so we only want to try this once (see above)
        sock_classes = []
        for fd in systemd_daemon.listen_fds():
            for family in (socket.AF_INET6, socket.AF_INET):
                if not systemd_daemon.is_socket_inet(
                    fd, family=family, type=socket.SOCK_DGRAM
                ):
                    continue
                sock = socket.fromfd(fd, family, socket.SOCK_DGRAM)
                sock_class = SocketClass(sock)
                sock_class.is_systemd = True
                sock_class.is_listener = True
                sock_classes.append(sock_class)
                self.sock_classes.append(sock_class)
                self.poller.register(sock_class, selectors.EVENT_READ)
                self.logger.debug("Opened socket: {}".format(sock_class))

        return sock_classes

    def setup_interface_address(
        self,
        interface_address,
        port=0,
        family=socket.AF_UNSPEC,
        type=socket.SOCK_DGRAM,
        proto=socket.IPPROTO_UDP,
        tags=None,
    ):
        sock_classes = []
        for addrinfo in socket.getaddrinfo(
            interface_address, port, family, type, proto
        ):
            if (
                (addrinfo[0] == socket.AF_INET6)
                and (not self.args.ipv4)
                and self.has_ipv6
            ):
                pass
            elif (addrinfo[0] == socket.AF_INET) and (not self.args.ipv6):
                pass
            else:
                continue

            if tags is not None and "listener" in tags:
                found_existing = False
                for sock_class in self.sock_classes_open:
                    if addrinfo == sock_class.bind_addrinfo:
                        sock_classes.append(sock_class)
                        found_existing = True
                if found_existing:
                    continue

            sock = self.new_socket(addrinfo)
            sock_class = SocketClass(sock)
            if tags is not None:
                for tag in tags:
                    setattr(sock_class, "is_{}".format(tag), True)
            sock_class.bind_addrinfo = addrinfo
            sock_classes.append(sock_class)
            self.sock_classes.append(sock_class)
            self.poller.register(sock_class, selectors.EVENT_READ)
            self.logger.debug("Opened socket: {}".format(sock_class))

        return sock_classes

    def get_system_addresses(self):
        if isinstance(netifaces, ImportError):
            return []

        addrs = set()
        for iface in netifaces.interfaces():
            iface_addrs = netifaces.ifaddresses(iface)
            if (self.args.ipv4 or (not self.args.ipv4 and not self.args.ipv6)) and (
                netifaces.AF_INET in iface_addrs
            ):
                addrs.update(
                    [
                        (f["addr"], netifaces.AF_INET)
                        for f in iface_addrs[netifaces.AF_INET]
                        if "addr" in f
                    ]
                )
            if (
                self.has_ipv6
                and (self.args.ipv6 or (not self.args.ipv4 and not self.args.ipv6))
                and (netifaces.AF_INET6 in iface_addrs)
            ):
                addrs.update(
                    [
                        (f["addr"], netifaces.AF_INET6)
                        for f in iface_addrs[netifaces.AF_INET6]
                        if "addr" in f
                    ]
                )

        return list(addrs)

    def set_high_port(self):
        interface_addresses = self.get_interface_addresses()
        interface_address = interface_addresses[0][0]
        family = interface_addresses[0][1]
        caught_errors = []
        for attempt in range(100):
            port = random.randint(49152, 65535)
            addrinfo = socket.getaddrinfo(
                interface_address, port, family, socket.SOCK_DGRAM, socket.IPPROTO_UDP
            )[0]
            sock = socket.socket(addrinfo[0], addrinfo[1], addrinfo[2])
            try:
                sock.bind(addrinfo[4])
            except socket.error as e:
                caught_errors.append((port, e))
                continue
            sock.close()
            if attempt > 0:
                self.logger.warning(
                    "It took {} attempts to get an unused port: {}".format(
                        attempt + 1, caught_errors
                    )
                )
            self.args.port = str(port)
            return
        raise caught_errors[-1][1]

    def get_interface_addresses(self):
        if self.args.interface_address:
            # Addresses supplied by user
            interface_addresses = [
                (addr, socket.AF_UNSPEC) for addr in self.args.interface_address
            ]
        elif not isinstance(netifaces, ImportError):
            # Addresses provided by netifaces
            interface_addresses = self.get_system_addresses()
        elif self.args.all_interfaces:
            # --all-interfaces is pretty much deprecated; this section is
            # only triggered if it's explicitly passed but netifaces is not
            # installed.
            raise socket.error(
                "All interface addresses not available; please install netifaces"
            )
        else:
            # Last resort
            interface_addresses = [("0.0.0.0", socket.AF_INET)]
            if self.has_ipv6:
                interface_addresses.append(("::", socket.AF_INET6))

        return interface_addresses

    def setup_sockets(self):
        if self.args.port == "-1":
            # Special testing mode
            self.set_high_port()

        if self.args.loopback:
            self.setup_sockets_loopback()
        if self.args.host:
            self.setup_sockets_client()
        if self.args.listen:
            self.gather_systemd_socks()
            self.setup_sockets_listener()

    def setup_sockets_loopback(self):
        # Client socket setup is only done once.
        if [
            sock_class
            for sock_class in self.sock_classes_open
            if sock_class.is_loopback and sock_class.is_client
        ]:
            return

        for i in range(self.args.loopback_pairs):
            sock_client, sock_listener = socket.socketpair(
                socket.AF_UNIX, socket.SOCK_DGRAM
            )
            sock_class_client = SocketClass(sock_client)
            sock_class_client.is_loopback = True
            sock_class_client.is_client = True
            sock_class_client.next_send = self.time_start
            self.poller.register(sock_class_client, selectors.EVENT_READ)
            self.sock_classes.append(sock_class_client)
            self.logger.debug("Opened socket: {}".format(sock_class_client))

            sock_class_listener = SocketClass(sock_listener)
            sock_class_listener.is_loopback = True
            sock_class_listener.is_listener = True
            self.poller.register(sock_class_listener, selectors.EVENT_READ)
            self.sock_classes.append(sock_class_listener)
            self.logger.debug("Opened socket: {}".format(sock_class_listener))

    def setup_sockets_listener(self):
        # Do not set up listener sockets if systemd sockets exist.
        if [
            sock_class for sock_class in self.sock_classes_open if sock_class.is_systemd
        ]:
            return

        interface_addresses = self.get_interface_addresses()
        interface_sock_classes = []
        for interface_address, interface_family in interface_addresses:
            interface_sock_classes += self.setup_interface_address(
                interface_address,
                port=self.args.port,
                family=interface_family,
                tags=["inet", "listener"],
            )

        listener_sock_classes = [
            sock_class
            for sock_class in self.sock_classes_open
            if sock_class.is_inet and sock_class.is_listener
        ]
        to_close = []
        for sock_class in listener_sock_classes:
            if sock_class not in interface_sock_classes:
                to_close.append(sock_class)

        self.close_socks(to_close)

    def get_srv_hosts(self):
        if isinstance(dns_resolver, ImportError):
            raise socket.error(
                "DNS SRV lookups not available; please install dnspython"
            )
        hosts = []
        for lookup in self.args.host:
            lookup_hosts_found = 0
            self.logger.debug("SRV lookup: {}".format(lookup))
            try:
                res = dns_resolver.query(
                    "_{}._udp.{}".format(self.args.srv_service, lookup), "srv"
                )
            except dns_resolver.dns.exception.DNSException as e:
                raise socket.error("{}: {}".format(lookup, e))
            for rdata in res:
                self.logger.debug("SRV result for {}: {}".format(lookup, rdata))
                if (str(rdata.target), rdata.port) in hosts:
                    continue
                hosts.append((str(rdata.target), rdata.port))
                lookup_hosts_found += 1
            if lookup_hosts_found == 0:
                raise socket.error("{}: No SRV results".format(lookup))
        return hosts

    def setup_sockets_client(self):
        # Client socket setup is only done once.
        if [
            sock_class
            for sock_class in self.sock_classes_open
            if sock_class.is_inet and sock_class.is_client
        ]:
            return

        if self.args.srv:
            hosts = self.get_srv_hosts()
        else:
            hosts = [(x, self.args.port) for x in self.args.host]
        for (hostname, port) in hosts:
            if str(port) in ("None", "-1"):
                port = self.args.port
            try:
                self.setup_sockets_client_host(hostname, port)
            except socket.error as e:
                eargs = list(e.args)
                if len(eargs) == 1:
                    eargs[0] = "{}: {}".format(hostname, eargs[0])
                else:
                    eargs[1] = "{}: {}".format(hostname, eargs[1])
                raise socket.error(*eargs)

    def setup_sockets_client_host(self, hostname, port):
        host_info = None
        for addrinfo in socket.getaddrinfo(
            hostname,
            port,
            socket.AF_UNSPEC,
            socket.SOCK_DGRAM,
            socket.IPPROTO_UDP,
            socket.AI_CANONNAME,
        ):
            if (
                (addrinfo[0] == socket.AF_INET6)
                and (not self.args.ipv4)
                and self.has_ipv6
            ):
                host_info = addrinfo
                break
            elif (addrinfo[0] == socket.AF_INET) and (not self.args.ipv6):
                host_info = addrinfo
                break
            else:
                continue
        if host_info is None:
            raise socket.error("Name or service not known")

        if len(self.args.interface_address) == 1:
            interface_address = self.args.interface_address[0]
        elif host_info[0] == socket.AF_INET6:
            interface_address = "::"
        else:
            interface_address = "0.0.0.0"

        for sock_class in self.setup_interface_address(
            interface_address,
            family=host_info[0],
            type=host_info[1],
            proto=host_info[2],
            tags=["inet", "client"],
        ):
            sock_class.client_host = host_info
            sock_class.next_send = self.time_start

    def send_new_ping(self, sock_class, peer_address=None):
        sock = sock_class.sock
        socket_address = sock_class.address
        if not peer_address and sock_class.client_host:
            peer_address = sock_class.client_host[4]
        peer_tuple = (socket_address, peer_address, sock.type)
        if peer_tuple not in sock_class.peer_states:
            sock_class.peer_states[peer_tuple] = PeerState(peer_tuple, sock_class)
        peer_state = sock_class.peer_states[peer_tuple]
        peer_state.last_seen = clock()

        packet_out = self.base_packet()
        packet_out.opcodes[
            packets.OpcodeReplyRequested.id
        ] = packets.OpcodeReplyRequested()
        self.start_investigations(peer_state, packet_out)
        dump_out = packet_out.dump()

        # If enabled, encrypt the packet and wrap it in a stub packet.
        if self.args.encrypt:
            encrypted_packet = packets.Packet()
            encrypted_packet.opcodes[
                packets.OpcodeEncrypted.id
            ] = packets.OpcodeEncrypted()
            encrypted_packet.opcodes[
                packets.OpcodeEncrypted.id
            ].method_index = self.args.encrypt_method_index
            encrypted_packet.opcodes[
                packets.OpcodeEncrypted.id
            ].session = sock_class.session
            encrypted_packet.opcodes[packets.OpcodeEncrypted.id].encrypt(
                dump_out, self.args.encrypt.encode("UTF-8")
            )
            sock_out = encrypted_packet.dump()
        else:
            sock_out = dump_out

        now = clock()
        self.sock_sendto(sock_class, sock_out, peer_address)
        sock_class.packets_transmitted += 1
        sock_class.pings_transmitted += 1
        peer_state.ping_position += 1
        peer_state.sent_messages[packet_out.message_id] = (
            now,
            packet_out.message_id,
            peer_state.ping_position,
        )
        packet_out_examine = packets.Packet()
        packet_out_examine.load(dump_out)
        if self.args.verbose:
            self.logger.info(
                "SEND{}: {}".format(
                    (" (encrypted)" if self.args.encrypt else ""), packet_out_examine
                )
            )

    def update_rtts(self, sock_class, rtt):
        for c in (sock_class,):
            c.rtt_total += rtt
            c.rtt_total_sq += rtt ** 2
            c.rtt_count += 1
            if (rtt < c.rtt_min) or (c.rtt_min == 0):
                c.rtt_min = rtt
            if rtt > c.rtt_max:
                c.rtt_max = rtt
            if c.rtt_ewma == 0:
                c.rtt_ewma = rtt * 8.0
            else:
                c.rtt_ewma += rtt - (c.rtt_ewma / 8.0)

    def sigquit_handler(self, signum, frame):
        self.print_stats(short=True)

    def sighup_handler(self, signum, frame):
        self.logger.debug("Received SIGHUP, scheduling reload")
        self.is_reload = True

    def get_nagios_stats(self):
        sock_class = [
            sock_class for sock_class in self.sock_classes if sock_class.is_client
        ][0]

        pings_lost = sock_class.pings_transmitted - sock_class.pings_received
        lost_pct = div0(pings_lost, sock_class.pings_transmitted) * 100
        rtt_avg = div0(float(sock_class.rtt_total), sock_class.rtt_count)

        if (lost_pct >= self.args.nagios_crit_loss) or (
            rtt_avg >= self.args.nagios_crit_rta
        ):
            nagios_result = 2
            nagios_result_text = "CRITICAL"
        elif (lost_pct >= self.args.nagios_warn_loss) or (
            rtt_avg >= self.args.nagios_warn_rta
        ):
            nagios_result = 1
            nagios_result_text = "WARNING"
        else:
            nagios_result = 0
            nagios_result_text = "OK"

        nagios_stats_text = _(
            "{result} 2PING - Packet loss = {loss}%, RTA = {avg:0.03f} ms"
        ).format(result=nagios_result_text, loss=int(lost_pct), avg=rtt_avg) + (
            "|rta={avg:0.06f}ms;{avgwarn:0.06f};{avgcrit:0.06f};0.000000 pl={loss}%;{losswarn};{losscrit};0"
        ).format(
            avg=rtt_avg,
            loss=int(lost_pct),
            avgwarn=self.args.nagios_warn_rta,
            avgcrit=self.args.nagios_crit_rta,
            losswarn=int(self.args.nagios_warn_loss),
            losscrit=int(self.args.nagios_crit_loss),
        )
        return nagios_result, nagios_stats_text

    def print_stats(self, short=False):
        time_end = clock()
        for sock_class in self.sock_classes:
            self.print_stats_sock(sock_class, time_end, short=short)

    def print_stats_sock(self, sock_class, time_end, short=False):
        stats_class = sock_class
        time_start = self.time_start
        pings_lost = stats_class.pings_transmitted - stats_class.pings_received
        lost_pct = div0(pings_lost, stats_class.pings_transmitted) * 100
        lost_undetermined = pings_lost - (
            stats_class.lost_outbound + stats_class.lost_inbound
        )
        outbound_pct = (
            div0(stats_class.lost_outbound, stats_class.pings_transmitted) * 100
        )
        inbound_pct = (
            div0(stats_class.lost_inbound, stats_class.pings_transmitted) * 100
        )
        undetermined_pct = div0(lost_undetermined, stats_class.pings_transmitted) * 100
        rtt_avg = div0(float(stats_class.rtt_total), stats_class.rtt_count)
        rtt_ewma = stats_class.rtt_ewma / 8.0
        rtt_mdev = math.sqrt(
            div0(stats_class.rtt_total_sq, stats_class.rtt_count)
            - (div0(stats_class.rtt_total, stats_class.rtt_count) ** 2)
        )
        socket_address = sock_class.address
        if sock_class.client_host:
            if sock_class.client_host[3]:
                hostname = "{} ({})".format(
                    sock_class.client_host[3], sock_class.client_host[4][0]
                )
            else:
                hostname = sock_class.client_host[4][0]
        elif socket_address:
            hostname = socket_address[0]
        else:
            hostname = sock_class
        if short:
            self.logger.info(
                _pl(
                    (
                        "{hostname}: {transmitted}/{received} ping, {loss}% loss "
                        "({outbound}/{inbound}/{undetermined} out/in/undet), min/avg/ewma/max/mdev = "
                        "{min:0.03f}/{avg:0.03f}/{ewma:0.03f}/{max:0.03f}/{mdev:0.03f} ms"
                    ),
                    (
                        "{hostname}: {transmitted}/{received} pings, {loss}% loss "
                        "({outbound}/{inbound}/{undetermined} out/in/undet), min/avg/ewma/max/mdev = "
                        "{min:0.03f}/{avg:0.03f}/{ewma:0.03f}/{max:0.03f}/{mdev:0.03f} ms"
                    ),
                    stats_class.pings_received,
                ).format(
                    hostname=hostname,
                    transmitted=stats_class.pings_transmitted,
                    received=stats_class.pings_received,
                    loss=int(lost_pct),
                    outbound=stats_class.lost_outbound,
                    inbound=stats_class.lost_inbound,
                    undetermined=lost_undetermined,
                    min=stats_class.rtt_min,
                    avg=rtt_avg,
                    ewma=rtt_ewma,
                    max=stats_class.rtt_max,
                    mdev=rtt_mdev,
                )
            )
        else:
            self.logger.info("")
            self.logger.info(
                "--- {} ---".format(
                    _("{hostname} 2ping statistics").format(hostname=hostname)
                )
            )
            self.logger.info(
                _pl(
                    "{transmitted} ping transmitted, {received} received, {loss}% ping loss, time {time}",
                    "{transmitted} pings transmitted, {received} received, {loss}% ping loss, time {time}",
                    stats_class.pings_transmitted,
                ).format(
                    transmitted=stats_class.pings_transmitted,
                    received=stats_class.pings_received,
                    loss=int(lost_pct),
                    time=stats_time(time_end - time_start),
                )
            )
            self.logger.info(
                _pl(
                    (
                        "{outbound} outbound ping loss ({outboundpct}%), {inbound} inbound ({inboundpct}%), "
                        "{undetermined} undetermined ({undeterminedpct}%)"
                    ),
                    (
                        "{outbound} outbound ping losses ({outboundpct}%), {inbound} inbound ({inboundpct}%), "
                        "{undetermined} undetermined ({undeterminedpct}%)"
                    ),
                    stats_class.lost_outbound,
                ).format(
                    outbound=stats_class.lost_outbound,
                    outboundpct=int(outbound_pct),
                    inbound=stats_class.lost_inbound,
                    inboundpct=int(inbound_pct),
                    undetermined=lost_undetermined,
                    undeterminedpct=int(undetermined_pct),
                )
            )
            self.logger.info(
                _(
                    "rtt min/avg/ewma/max/mdev = {min:0.03f}/{avg:0.03f}/{ewma:0.03f}/{max:0.03f}/{mdev:0.03f} ms"
                ).format(
                    min=stats_class.rtt_min,
                    avg=rtt_avg,
                    ewma=rtt_ewma,
                    max=stats_class.rtt_max,
                    mdev=rtt_mdev,
                )
            )
            self.logger.info(
                _pl(
                    "{transmitted} raw packet transmitted, {received} received",
                    "{transmitted} raw packets transmitted, {received} received",
                    stats_class.packets_transmitted,
                ).format(
                    transmitted=stats_class.packets_transmitted,
                    received=stats_class.packets_received,
                )
            )

    def ready(self):
        pass

    def run(self):
        self.logger.debug("Clock value: {:f}".format(clock()))
        self.logger.debug("Poller: {}".format(self.poller))
        if hasattr(signal, "SIGQUIT"):
            signal.signal(signal.SIGQUIT, self.sigquit_handler)
        if hasattr(signal, "SIGHUP"):
            signal.signal(signal.SIGHUP, self.sighup_handler)

        try:
            self.setup_sockets()
        except (socket.error, socket.gaierror) as e:
            if self.args.debug:
                raise
            else:
                self.logger.error(str(e))
                return 1

        for sock_class in self.sock_classes_active:
            if self.args.nagios:
                continue
            socket_address = sock_class.address
            if sock_class.client_host:
                if sock_class.client_host[3]:
                    host_display = "{} ({})".format(
                        sock_class.client_host[3], sock_class.client_host[4][0]
                    )
                else:
                    host_display = "{}".format(sock_class.client_host[4][0])
            elif sock_class.next_send:
                host_display = "{}".format(sock_class)
            elif socket_address:
                host_display = "listener ({})".format(socket_address[0])
            else:
                host_display = "listener ({})".format(sock_class)
            self.logger.info(
                _("2PING {host}: {min} to {max} bytes of data.").format(
                    host=host_display,
                    min=self.args.min_packet_size,
                    max=self.args.max_packet_size,
                )
            )

        self.ready()

        if self.args.preload > 1:
            for sock_class in self.sock_classes_active:
                if not sock_class.next_send:
                    continue
                for i in range(self.args.preload - 1):
                    self.send_new_ping(sock_class)

        try:
            self.loop()
        except KeyboardInterrupt:
            pass

        if self.args.nagios:
            nagios_result, nagios_stats_text = self.get_nagios_stats()
            self.print_out(nagios_stats_text)
        else:
            self.print_stats()
        self.close_socks()
        if self.args.nagios:
            return nagios_result
        else:
            return 0

    def base_packet(self):
        packet_out = packets.Packet()
        if (not self.args.no_send_version) or (self.args.notice):
            packet_out.opcodes[packets.OpcodeExtended.id] = packets.OpcodeExtended()
        if not self.args.no_send_version:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedVersion.id
            ] = packets.ExtendedVersion()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedVersion.id
            ].text = version_string
        if self.args.send_time:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedWallClock.id
            ] = packets.ExtendedWallClock()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedWallClock.id
            ].time_us = int(time.time() * 1000000)
        if self.args.send_monotonic_clock:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedMonotonicClock.id
            ] = packets.ExtendedMonotonicClock()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedMonotonicClock.id
            ].generation = self.fake_time_generation
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedMonotonicClock.id
            ].time_us = int(
                (clock() - self.time_start + self.fake_time_epoch) * 1000000
            )
        if self.args.send_random:
            random_data = bytes(
                [random.randint(0, 255) for x in range(self.args.send_random)]
            )
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedRandom.id
            ] = packets.ExtendedRandom()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedRandom.id
            ].is_hwrng = False
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedRandom.id
            ].is_os = random_is_systemrandom
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedRandom.id
            ].random_data = random_data
        if self.args.notice:
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedNotice.id
            ] = packets.ExtendedNotice()
            packet_out.opcodes[packets.OpcodeExtended.id].segments[
                packets.ExtendedNotice.id
            ].text = self.args.notice
        if self.args.auth:
            packet_out.opcodes[packets.OpcodeHMAC.id] = packets.OpcodeHMAC()
            packet_out.opcodes[packets.OpcodeHMAC.id].key = self.args.auth.encode(
                "UTF-8"
            )
            packet_out.opcodes[
                packets.OpcodeHMAC.id
            ].digest_index = self.args.auth_digest_index
        packet_out.padding_pattern = self.args.pattern_bytes
        packet_out.min_length = self.args.min_packet_size
        packet_out.max_length = self.args.max_packet_size
        if self.args.encrypt:
            packet_out.align_length = 16
        return packet_out

    def scheduled_cleanup(self):
        self.logger.debug("Cleanup")
        self.setup_sockets()
        for sock_class in self.sock_classes_active:
            self.scheduled_cleanup_sock_class(sock_class)

    def scheduled_cleanup_sock_class(self, sock_class):
        now = clock()
        for peer_tuple in tuple(sock_class.peer_states.keys()):
            peer_state = sock_class.peer_states[peer_tuple]
            if now > peer_state.last_seen + 600.0:
                del sock_class.peer_states[peer_tuple]
                self.logger.debug("Cleanup: Removed {}".format(peer_tuple))
                continue
            for table_name, max_time in (
                ("sent_messages", 600.0),
                ("seen_messages", 600.0),
                ("courtesy_messages", 120.0),
                ("encrypted_session_ivs", 600.0),
            ):
                table = getattr(peer_state, table_name)
                for table_key_name in tuple(table.keys()):
                    if now > (table[table_key_name][0] + max_time):
                        del table[table_key_name]
                        self.logger.debug(
                            "Cleanup: Removed {} {} {}".format(
                                peer_tuple, table_name, table_key_name
                            )
                        )

    def new_socket(self, addrinfo):
        family = addrinfo[0]
        type = addrinfo[1]
        proto = addrinfo[2]
        bind = addrinfo[4]

        sock = socket.socket(family, type, proto)
        try:
            import IN

            sock.setsockopt(socket.IPPROTO_IP, IN.IP_RECVERR, int(True))
        except (ImportError, AttributeError, socket.error):
            pass
        if family == socket.AF_INET6:
            try:
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, int(True))
            except (AttributeError, socket.error):
                pass
        sock.bind(bind)
        return sock

    def get_next_wakeup(self):
        now = clock()
        events = [(self.next_cleanup, "cleanup")]
        for sock_class in self.sock_classes_active:
            if sock_class.next_send:
                events.append((sock_class.next_send, "send ({})".format(sock_class)))
            if sock_class.shutdown_time > now:
                events.append(
                    (sock_class.shutdown_time, "shutdown ({})".format(sock_class))
                )
        if self.args.stats:
            events.append((self.next_stats, "stats"))
        if self.args.deadline:
            events.append((self.time_start + self.args.deadline, "deadline"))

        next_wakeup = now + self.old_age_interval
        next_wakeup_reason = "old age"
        for wakeup, reason in events:
            if wakeup < next_wakeup:
                next_wakeup = wakeup
                next_wakeup_reason = reason
        if next_wakeup < now:
            next_wakeup = now
            next_wakeup_reason = "time travel ({})".format(next_wakeup_reason)

        return next_wakeup, next_wakeup_reason

    def schedule_next_send(self, sock_class, reply_received=False):
        now = clock()
        if self.args.adaptive and sock_class.rtt_ewma:
            # Adaptive gets recalculated immediately.
            sock_class.next_send = now + (sock_class.rtt_ewma / 8.0 / 1000.0)
        elif self.args.flood:
            if reply_received:
                # If we're in flood mode and a ping reply was received, send a new ping ASAP.
                sock_class.next_send = now
            else:
                # Send has just happened, give it a short time for reply.
                sock_class.next_send = now + 0.01
        elif not reply_received:
            sock_class.next_send = now + self.args.interval

    def loop_client_send(self):
        now = clock()
        for sock_class in self.sock_classes_active:
            if not sock_class.next_send:
                # Not scheduled to send
                continue
            if now < sock_class.next_send:
                # Not yet time
                continue

            if self.args.count and (sock_class.pings_transmitted >= self.args.count):
                self.logger.debug(
                    "loop_client_send: Setting shutdown time to +{} on {}".format(
                        self.args.interval, sock_class
                    )
                )
                sock_class.shutdown_time = now + self.args.interval
                sock_class.next_send = 0
                continue
            self.send_new_ping(sock_class)
            self.schedule_next_send(sock_class)

    def loop_poller(self, next_wakeup):
        timeout = next_wakeup - clock()
        for key, mask in self.poller.select(timeout=(0 if timeout < 0 else timeout)):
            sock_class = key.fileobj
            try:
                self.process_incoming_packet(sock_class)
            except Exception:
                self.logger.exception(
                    _("Received unexpected exception (please file a bug report)")
                )
                if self.args.debug:
                    raise

            if (
                self.args.count
                and (sock_class.pings_transmitted >= self.args.count)
                and (sock_class.pings_transmitted == sock_class.pings_received)
            ):
                self.logger.debug(
                    "loop_poller: Setting shutdown time to now on {}".format(sock_class)
                )
                sock_class.shutdown_time = clock()
                sock_class.next_send = 0

    def loop(self):
        while True:
            self.loop_client_send()
            next_wakeup, next_wakeup_reason = self.get_next_wakeup()
            self.logger.debug(
                "Next wakeup: {} ({})".format(
                    (next_wakeup - clock()), next_wakeup_reason
                )
            )
            self.loop_poller(next_wakeup)

            now = clock()

            if self.args.stats and now >= self.next_stats:
                self.print_stats(short=True)
                self.next_stats = now + self.args.stats

            if self.args.deadline and now >= (self.time_start + self.args.deadline):
                for sock_class in self.sock_classes_active:
                    self.logger.debug(
                        "loop (deadline): Setting shutdown time to now on {}".format(
                            sock_class
                        )
                    )
                    sock_class.shutdown_time = now
                    sock_class.next_send = 0

            if (
                self.args.count
                and self.args.no_3way
                and len(
                    [
                        sock_class
                        for sock_class in self.sock_classes_active
                        if sock_class.is_client
                    ]
                )
                == 0
            ):
                # Special case for --count and --no-3way on a listener
                for sock_class in [
                    sock_class
                    for sock_class in self.sock_classes_active
                    if sock_class.is_listener
                ]:
                    self.logger.debug(
                        "loop (no_3way): Setting shutdown time to now on {}".format(
                            sock_class
                        )
                    )
                    sock_class.shutdown_time = clock()
                    sock_class.next_send = 0

            if now >= self.next_cleanup:
                self.scheduled_cleanup()
                self.next_cleanup = now + 60.0

            if self.is_reload:
                self.is_reload = False
                self.setup_sockets()

            if len(self.sock_classes_active) == 0:
                return


def main(argv=None):
    if argv is None:
        argv = sys.argv

    args = parse_args(argv)
    t = TwoPing(args)

    if t.args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    logging.basicConfig(format="%(message)s", stream=sys.stdout, level=log_level)

    return t.run()


def module_init():
    if __name__ == "__main__":
        sys.exit(main(sys.argv))


module_init()



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2021 Ryan Finnie
# SPDX-License-Identifier: GPL-2.0+

import hashlib
import hmac
from math import ceil
import time

from . import crc32
from .utils import AES, npack, nunpack, random, twoping_checksum


class Extended:
    id = None

    def __init__(self):
        self.data = b""

    def __repr__(self):
        if self.id is None:
            return "<Extended: {} bytes>".format(len(self.data))
        else:
            return "<Extended (0x{:08x}): {} bytes>".format(self.id, len(self.data))

    def load(self, data):
        self.data = data

    def dump(self):
        return self.data


class ExtendedText(Extended):
    def __init__(self):
        self.text = str()

    def __repr__(self):
        return "<Text (Generic): {}>".format(self.text)

    def load(self, data):
        self.text = data.decode("UTF-8")

    def dump(self, max_length=None):
        text_bytes = self.text.encode("UTF-8")
        if (max_length is not None) and (max_length < len(text_bytes)):
            return None
        return text_bytes


class ExtendedVersion(ExtendedText):
    id = 0x3250564E

    def __repr__(self):
        return "<Version: {}>".format(self.text)


class ExtendedNotice(ExtendedText):
    id = 0xA837B44E

    def __repr__(self):
        return "<Notice: {}>".format(self.text)


class ExtendedWallClock(Extended):
    id = 0x64F69319

    def __init__(self):
        self.time_us = 0

    def __repr__(self):
        return "<Wall Clock: {}>".format(
            time.strftime("%c", time.gmtime(self.time_us / 1000000.0))
        )

    def load(self, data):
        self.time_us = nunpack(data[0:8])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 8):
            return None
        return npack(self.time_us, 8)


class ExtendedMonotonicClock(Extended):
    id = 0x771D8DFB

    def __init__(self):
        self.generation = 0
        self.time_us = 0

    def __repr__(self):
        return "<Monotonic Clock: {:0.9f}, gen {}>".format(
            (self.time_us / 1000000.0), self.generation
        )

    def load(self, data):
        self.generation = nunpack(data[0:2])
        self.time_us = nunpack(data[2:10])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 10):
            return None
        return npack(self.generation, 2) + npack(self.time_us, 8)


class ExtendedRandom(Extended):
    id = 0x2FF6AD68

    def __init__(self):
        self.is_hwrng = False
        self.is_os = False
        self.random_data = b""

    def __repr__(self):
        return "<Random: {} ({}), HWRNG {}, OS {}>".format(
            repr(self.random_data),
            len(self.random_data),
            repr(self.is_hwrng),
            repr(self.is_os),
        )

    def load(self, data):
        flags = nunpack(data[0:2])
        self.is_hwrng = bool(flags & 0x0001)
        self.is_os = bool(flags & 0x0002)
        self.random_data = data[2:]

    def dump(self, max_length=None):
        random_data = self.random_data
        if len(random_data) == 0:
            return None
        if max_length is not None:
            if max_length < 3:
                return None
            if max_length < (len(random_data) - 2):
                random_data = random_data[0 : max_length - 2]
        flags = 0
        if self.is_hwrng:
            flags = flags | 0x0001
        if self.is_os:
            flags = flags | 0x0002
        return npack(flags, 2) + random_data


class ExtendedBatteryLevels(Extended):
    id = 0x88A1F7C7

    def __init__(self):
        self.batteries = {}

    def __repr__(self):
        return "<Batteries ({}): [{}]>".format(
            len(self.batteries),
            ", ".join(
                [
                    "{}: {:0.03%}".format(x, self.batteries[x] / 65535.0)
                    for x in sorted(self.batteries)
                ]
            ),
        )

    def load(self, data):
        self.batteries = {}
        pos = 2
        for i in range(nunpack(data[0:2])):
            battery_id = nunpack(data[pos : pos + 2])
            battery_level = nunpack(data[pos + 2 : pos + 4])
            self.batteries[battery_id] = battery_level
            pos += 4

    def dump(self, max_length=None):
        if max_length is not None:
            if max_length < 6:
                return None
            batteries = {}
            for i in sorted(self.batteries.keys())[0 : int((max_length - 2) / 4)]:
                batteries[i] = self.batteries[i]
        else:
            batteries = self.batteries

        out = npack(len(batteries), 2)
        for i in batteries:
            out += npack(i, 2)
            out += npack(batteries[i], 2)
        return out


class Opcode:
    id = None

    def __init__(self):
        self.data = b""

    def __repr__(self):
        if self.id is None:
            return "<Opcode: {} bytes>".format(len(self.data))
        else:
            return "<Opcode (0x{:04x}): {} bytes>".format(self.id, len(self.data))

    def load(self, data):
        self.data = data

    def dump(self):
        return self.data


class OpcodeReplyRequested(Opcode):
    id = 0x0001

    def __init__(self):
        pass

    def __repr__(self):
        return "<Reply Requested>"

    def load(self, data):
        pass

    def dump(self, max_length=None):
        return b""


class OpcodeInReplyTo(Opcode):
    id = 0x0002

    def __init__(self):
        self.message_id = b""

    def __repr__(self):
        return "<In Reply To: 0x{}>".format(self.message_id.hex())

    def load(self, data):
        self.message_id = data[0:6]

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 6):
            return None
        return self.message_id


class OpcodeRTTEnclosed(Opcode):
    id = 0x0004

    def __init__(self):
        self.rtt_us = 0

    def __repr__(self):
        return "<RTT Enclosed: {} us>".format(self.rtt_us)

    def load(self, data):
        self.rtt_us = nunpack(data[0:4])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 4):
            return None
        return npack(self.rtt_us, 4)


class OpcodeMessageIDList(Opcode):
    _repr_name = "ID List (Generic)"

    def __init__(self):
        self.message_ids = []

    def __repr__(self):
        return "<{}: [{}] ({})>".format(
            self._repr_name,
            ", ".join(["0x{}".format(x.hex()) for x in self.message_ids]),
            len(self.message_ids),
        )

    def load(self, data):
        self.message_ids = []
        pos = 2
        for i in range(nunpack(data[0:2])):
            self.message_ids.append(data[pos : pos + 6])
            pos += 6

    def dump(self, max_length=None):
        if max_length is not None:
            if max_length < 8:
                return None
            output_ids = self.message_ids[0 : int((max_length - 2) / 6)]
        else:
            output_ids = self.message_ids

        out = npack(len(output_ids), 2)
        for i in output_ids:
            out += i
        return out


class OpcodeInvestigationSeen(OpcodeMessageIDList):
    id = 0x0008
    _repr_name = "Investigation Seen"


class OpcodeInvestigationUnseen(OpcodeMessageIDList):
    id = 0x0010
    _repr_name = "Investigation Unseen"


class OpcodeInvestigate(OpcodeMessageIDList):
    id = 0x0020
    _repr_name = "Investigate"


class OpcodeCourtesyExpiration(OpcodeMessageIDList):
    id = 0x0040
    _repr_name = "Courtesy Expiration"


class OpcodeHMAC(Opcode):
    id = 0x0080

    def __init__(self):
        self.key = b""
        self.digest_index = None
        self.hash = b""

        self.digest_map = {
            1: (hashlib.md5, 16, "HMAC-MD5"),
            2: (hashlib.sha1, 20, "HMAC-SHA1"),
            3: (hashlib.sha256, 32, "HMAC-SHA256"),
            4: (crc32, 4, "HMAC-CRC32"),
            5: (hashlib.sha512, 64, "HMAC-SHA512"),
        }

    def __repr__(self):
        if self.digest_index is not None:
            return "<{}: 0x{}>".format(
                self.digest_map[self.digest_index][2], self.hash.hex()
            )
        return "<HMAC>"

    def load(self, data):
        self.digest_index = nunpack(data[0:2])
        self.hash = data[2:]

    def dump(self, max_length=None):
        if self.digest_index is not None:
            (hasher, size, hasher_name) = self.digest_map[self.digest_index]
            return npack(self.digest_index, 2) + bytes(size)
        return None


class OpcodeHostLatency(Opcode):
    id = 0x0100

    def __init__(self):
        self.delay_us = 0

    def __repr__(self):
        return "<Host Latency: {} us>".format(self.delay_us)

    def load(self, data):
        self.delay_us = nunpack(data[0:4])

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 4):
            return None
        return npack(self.delay_us, 4)


class OpcodeEncrypted(Opcode):
    id = 0x0200

    def __init__(self):
        self.hkdf_info = (
            b"\xd8\x89\xac\x93\xac\xeb\xa1\xf3\x98\xd0\xc6\x9b\xc8\xc6\xa7\xaa"
        )
        self.method_index = None
        self.encrypted = b""
        self.session = b""
        self.iv = None

        self.method_map = {1: ("HKDF-AES256-CBC",)}

    def __repr__(self):
        if self.method_index is not None:
            return "<{} (Session {}, IV {}, {} bytes)>".format(
                self.method_map[self.method_index][0],
                repr(self.session),
                repr(self.iv),
                len(self.encrypted),
            )
        return "<Encrypted>"

    def load(self, data):
        self.method_index = nunpack(data[0:2])
        if self.method_index == 1:
            self.session = data[2:10]
            self.iv = data[10:26]
            self.encrypted = data[26:]
        else:
            self.encrypted = data[2:]

    def dump(self, max_length=None):
        if self.method_index == 1:
            return npack(self.method_index, 2) + self.session + self.iv + self.encrypted
        return None

    def encrypt(self, unencrypted, key):
        if isinstance(AES, ImportError):
            return None
        if self.method_index is None:
            return None
        if self.method_index == 1:
            if self.iv is None:
                self.iv = bytes([random.randint(0, 255) for x in range(16)])
            aeskey = self.hkdf(
                32,
                key,
                salt=self.iv,
                info=(self.hkdf_info + self.session),
                digestmod=hashlib.sha256,
            )
            aes_e = AES.new(aeskey, AES.MODE_CBC, self.iv)
            self.encrypted = aes_e.encrypt(unencrypted)
        else:
            return None

    def decrypt(self, key):
        if isinstance(AES, ImportError):
            return None
        if self.method_index is None:
            return None
        if self.method_index == 1:
            aeskey = self.hkdf(
                32,
                key,
                salt=self.iv,
                info=(self.hkdf_info + self.session),
                digestmod=hashlib.sha256,
            )
            aes_d = AES.new(aeskey, AES.MODE_CBC, self.iv)
            return aes_d.decrypt(self.encrypted)

    def hkdf(self, length, ikm, salt=b"", info=b"", digestmod=None):
        if digestmod is None:
            digestmod = hashlib.sha256
        prk = hmac.new(salt, ikm, digestmod).digest()
        hash_len = len(prk)
        t = b""
        okm = b""
        for i in range(ceil(length / hash_len)):
            t = hmac.new(prk, t + info + bytes([1 + i]), digestmod).digest()
            okm += t
        return okm[:length]


class OpcodeExtended(Opcode):
    id = 0x8000

    def __init__(self):
        self.segments = {}
        self.segment_data_positions = {}

    def __repr__(self):
        return "<Extended: {}>".format(
            repr(sorted(self.segments.values(), key=lambda x: x.id))
        )

    def load(self, data):
        self.segments = {}
        self.segment_data_positions = {}

        pos = 0
        known_segments = (
            ExtendedVersion,
            ExtendedNotice,
            ExtendedMonotonicClock,
            ExtendedWallClock,
            ExtendedRandom,
            ExtendedBatteryLevels,
        )

        while pos < len(data):
            flag = nunpack(data[pos : pos + 4])
            pos += 4
            segment_data_length = nunpack(data[pos : pos + 2])
            pos += 2
            self.segment_data_positions[flag] = (pos, segment_data_length)
            segment_handler = None
            for seg in known_segments:
                if flag == seg.id:
                    segment_handler = seg
                    break
            if segment_handler is None:
                segment_handler = Extended
                segment_handler.id = flag
            self.segments[flag] = segment_handler()
            self.segments[flag].load(data[pos : (pos + segment_data_length)])
            pos += segment_data_length

    def dump(self, max_length=None):
        if (max_length is not None) and (max_length < 6):
            return None
        out = b""
        pos = 0
        for segment in self.segments.values():
            if max_length is None:
                segment_max_length = None
            else:
                segment_max_length = max_length - pos - 6
            segment_data = segment.dump(max_length=segment_max_length)
            if segment_data is None:
                continue
            out += npack(segment.id, 4)
            pos += 4
            out += npack(len(segment_data), 2)
            pos += 2
            out += segment_data
            pos += len(segment_data)
        if len(out) == 0:
            return None
        return out


class Packet:
    def __repr__(self):
        return "<Packet (0x{}): {}>".format(
            self.message_id.hex(),
            repr(sorted(self.opcodes.values(), key=lambda x: x.id)),
        )

    def __init__(self):
        self.message_id = b""
        self.opcodes = {}
        self.min_length = 0
        self.max_length = 1024
        self.align_length = 0
        self.padding_pattern = b"\x00"
        self.opcode_data_positions = {}

    def load(self, data):
        magic_number = data[0:2]
        if magic_number != b"\x32\x50":
            raise Exception("Invalid magic number")
        checksum = nunpack(data[2:4])
        if checksum:
            if twoping_checksum(data[0:2] + b"\x00\x00" + data[4:]) != checksum:
                raise Exception("Invalid checksum")
        self.message_id = data[4:10]
        opcode_flags = nunpack(data[10:12])
        self.opcodes = {}

        pos = 12
        known_opcodes = (
            OpcodeReplyRequested,
            OpcodeInReplyTo,
            OpcodeRTTEnclosed,
            OpcodeInvestigationSeen,
            OpcodeInvestigationUnseen,
            OpcodeInvestigate,
            OpcodeCourtesyExpiration,
            OpcodeHMAC,
            OpcodeHostLatency,
            OpcodeEncrypted,
            OpcodeExtended,
        )
        for flag in (2 ** x for x in range(16)):
            if not opcode_flags & flag:
                continue
            opcode_data_length = nunpack(data[pos : pos + 2])
            pos += 2
            self.opcode_data_positions[flag] = (pos, opcode_data_length)
            opcode_handler = None
            for oc in known_opcodes:
                if flag == oc.id:
                    opcode_handler = oc
                    break
            if opcode_handler is None:
                opcode_handler = Opcode
                opcode_handler.id = flag
            self.opcodes[flag] = opcode_handler()
            self.opcodes[flag].load(data[pos : (pos + opcode_data_length)])
            pos += opcode_data_length

    def dump(self):
        auth_pos_begin = 0
        auth_pos_end = 0
        if not self.message_id:
            self.message_id = bytes([random.randint(0, 255) for x in range(6)])
        opcode_datas = {}
        packet_length = 12
        for flag in (
            OpcodeEncrypted.id,
            OpcodeHMAC.id,
            OpcodeReplyRequested.id,
            OpcodeInReplyTo.id,
            OpcodeRTTEnclosed.id,
            OpcodeInvestigationSeen.id,
            OpcodeInvestigationUnseen.id,
            OpcodeInvestigate.id,
            OpcodeHostLatency.id,
            OpcodeCourtesyExpiration.id,
            OpcodeExtended.id,
        ):
            if flag not in self.opcodes:
                continue
            if (packet_length + 2) > self.max_length:
                break
            res = self.opcodes[flag].dump(
                max_length=(self.max_length - packet_length - 2)
            )
            if res is None:
                continue
            opcode_datas[flag] = res
            res_len = len(res)
            packet_length += res_len + 2
        opcode_flags = 0
        opcode_data = b""
        packet_length = 12
        for flag in sorted(opcode_datas.keys()):
            res = opcode_datas[flag]
            res_len = len(res)
            if flag == OpcodeHMAC.id:
                auth_pos_begin = packet_length + 4
                auth_pos_end = auth_pos_begin + (res_len - 2)
            opcode_flags = opcode_flags | flag
            opcode_data += npack(res_len, 2)
            opcode_data += res
            packet_length += res_len + 2
        out = bytearray(
            b"\x32\x50\x00\x00" + self.message_id + npack(opcode_flags, 2) + opcode_data
        )
        target_length = len(out)
        if len(out) < self.min_length:
            target_length = self.min_length
        if self.align_length and (target_length % self.align_length):
            target_length += self.align_length - (target_length % self.align_length)
        if len(out) < target_length:
            target_padding = target_length - len(out)
            padding = (
                self.padding_pattern
                * int(target_padding / len(self.padding_pattern) + 1)
            )[0:target_padding]
            out += padding
        if (OpcodeHMAC.id in self.opcodes) and auth_pos_begin:
            out[auth_pos_begin:auth_pos_end] = self.calculate_hash(
                self.opcodes[OpcodeHMAC.id], out
            )
        out[2:4] = npack(twoping_checksum(out), 2)
        return bytes(out)

    def calculate_hash(self, opcode, payload):
        (hasher, size, hasher_name) = opcode.digest_map[opcode.digest_index]
        return hmac.new(opcode.key, payload, hasher).digest()



import hmac
import unittest

from . import _test_module_init
from twoping import crc32


class TestCRC32(unittest.TestCase):
    def test_crc32(self):
        c = crc32.new(b"Data to hash")
        self.assertEqual(c.digest(), b"\x44\x9e\x0a\x5c")

    def test_hmac(self):
        h = hmac.new(b"Secret key", b"Data to hash", crc32)
        self.assertEqual(h.digest(), b"\x3c\xe1\xb6\xb9")

    def test_update(self):
        c = crc32.new()
        c.update(b"Data to hash")
        self.assertEqual(c.digest(), b"\x44\x9e\x0a\x5c")

    def test_hexdigest(self):
        c = crc32.new(b"Data to hash")
        self.assertEqual(c.hexdigest(), "449e0a5c")

    def test_hexdigest_zero_padding(self):
        c = crc32.new(b"jade")
        self.assertEqual(c.hexdigest(), "00835218")

    def test_clear(self):
        c = crc32.new(b"Data to hash")
        c.clear()
        self.assertEqual(c.digest(), b"\x00\x00\x00\x00")

    def test_zero_padding(self):
        c = crc32.new(b"jade")
        self.assertEqual(c.digest(), b"\x00\x83\x52\x18")

    def test_module_init(self):
        self.assertTrue(_test_module_init(crc32))



import locale
import logging
import pytest
import unittest
import unittest.mock

from . import _test_module_init
from twoping import args, cli, utils


class TestCLI(unittest.TestCase):
    bind_addresses = ["127.0.0.1"]
    port = None
    logger = None
    class_args = None

    def setUp(self):
        self.logger = logging.getLogger()
        self.logger.level = logging.DEBUG

    def _client(self, test_flags=(), test_positionals=(), test_stats=True, pairs=1):
        if self.port is None:
            port = -1
        else:
            port = self.port

        flag_args = ["--debug"]
        if self.class_args is not None:
            flag_args += self.class_args
        flag_args += test_flags

        positional_args = list(test_positionals)

        if "--loopback" not in flag_args:
            flag_args += ["--listen", "--port={}".format(port)]
            for bind_address in self.bind_addresses:
                flag_args.append("--interface-address={}".format(bind_address))
                positional_args.append(bind_address)
        if ("--adaptive" not in flag_args) and ("--flood" not in flag_args):
            flag_args.append("--count=1")
        if not ("--count=1" in flag_args):
            flag_args.append("--interval=5")
        all_args = ["2ping"] + flag_args + positional_args
        self.logger.info("Passed arguments: {}".format(all_args))

        p = cli.TwoPing(args.parse_args(all_args))
        self.logger.info("Parsed arguments: {}".format(p.args))
        self.assertEqual(p.run(), 0)

        for sock_class in p.sock_classes:
            self.assertTrue(sock_class.closed)

        self.assertEqual(len(p.sock_classes), (pairs * 2))
        client_sock_classes = [
            sock_class for sock_class in p.sock_classes if sock_class.is_client
        ]
        self.assertEqual(len(client_sock_classes), pairs)
        sock_class = client_sock_classes[0]

        if not test_stats:
            return sock_class

        self.assertEqual(sock_class.errors_received, 0)
        self.assertEqual(sock_class.lost_inbound, 0)
        self.assertEqual(sock_class.lost_outbound, 0)
        if p.args.count:
            self.assertEqual(sock_class.pings_transmitted, p.args.count)
            self.assertEqual(sock_class.pings_received, p.args.count)

        return sock_class

    def test_3way(self):
        sock_class = self._client([])
        self.assertEqual(sock_class.packets_transmitted, 2)
        self.assertEqual(sock_class.packets_received, 1)

    def test_3way_no(self):
        sock_class = self._client(["--no-3way"])
        self.assertEqual(sock_class.packets_transmitted, 1)
        self.assertEqual(sock_class.packets_received, 1)

    @pytest.mark.slow
    def test_adaptive(self):
        sock_class = self._client(["--adaptive", "--deadline=3"])
        self.assertGreaterEqual(sock_class.pings_transmitted, 100)

    @unittest.skipIf(isinstance(utils.AES, ImportError), "Crypto module required")
    def test_encrypt(self):
        self._client(
            ["--encrypt-method=hkdf-aes256-cbc", "--encrypt=S49HVbnJd3fBdDzdMVVw"]
        )

    @pytest.mark.slow
    def test_flood(self):
        sock_class = self._client(["--flood", "--deadline=3"])
        self.assertGreaterEqual(sock_class.pings_transmitted, 100)

    def test_hmac_crc32(self):
        self._client(["--auth-digest=hmac-crc32", "--auth=mc82kJwtXFlhqQSCKptQ"])

    def test_hmac_md5(self):
        self._client(["--auth-digest=hmac-md5", "--auth=rBgRpBfRbF4DkwFQXncz"])

    def test_hmac_sha1(self):
        self._client(["--auth-digest=hmac-sha1", "--auth=qnzTCJHnZXdrxRZ8JjQw"])

    def test_hmac_sha256(self):
        self._client(["--auth-digest=hmac-sha256", "--auth=cc8G2Ssbq4WZRq7H7d5L"])

    def test_hmac_sha512(self):
        self._client(["--auth-digest=hmac-sha512", "--auth=sjk3kqzcSV3XfHJWNstn"])

    def test_invalid_hostname(self):
        with self.assertRaises(OSError):
            self._client([], ["xGkKWDDMnZxCD4XchMnK."])

    def test_monotonic_clock(self):
        self._client(["--send-monotonic-clock"])

    def test_notice(self):
        self._client(["--notice=Notice text"])

    @unittest.skipUnless(
        (locale.getlocale()[1] == "UTF-8"), "UTF-8 environment required"
    )
    def test_notice_utf8(self):
        self._client(["--notice=UTF-8 \u2603"])

    @pytest.mark.slow
    def test_packet_loss(self):
        # There is a small but non-zero chance that packet loss will prevent
        # any investigation replies from getting back to the client sock
        # between the 10 and 15 second mark, causing a test failure.
        # There's an even more minisculely small chance no simulated losses
        # will occur within the test period.
        sock_class = self._client(
            ["--flood", "--deadline=15", "--packet-loss=25"], test_stats=False
        )
        self.assertGreaterEqual(sock_class.pings_transmitted, 100)
        self.assertEqual(sock_class.errors_received, 0)
        self.assertGreater(sock_class.lost_inbound, 0)
        self.assertGreater(sock_class.lost_outbound, 0)

    def test_random(self):
        self._client(["--send-random=32"])

    def test_time(self):
        self._client(["--send-time"])

    def test_module_init(self):
        self.assertTrue(_test_module_init(cli))


class TestCLIInet(TestCLI):
    def test_srv_addresses(self):
        self.bind_addresses = []
        with unittest.mock.patch(
            "twoping.cli.TwoPing.get_srv_hosts", return_value=[("127.0.0.1", -1)]
        ):
            self._client(
                ["--interface-address=127.0.0.1", "--srv"], ["pssPCPkc3XlMTDZhclPV."]
            )


@unittest.skipUnless(hasattr(cli.socket, "AF_UNIX"), "UNIX environment required")
class TestCLILoopback(TestCLI):
    class_args = ["--loopback"]

    def test_loopback_pairs(self):
        self._client(["--loopback-pairs=3"], pairs=3)



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2021 Ryan Finnie
# SPDX-License-Identifier: GPL-2.0+

import gettext
import platform
import random as _pyrandom

try:
    import distro
except ImportError as e:
    distro = e

try:
    from Cryptodome.Cipher import AES
except ImportError as e:
    try:
        from Crypto.Cipher import AES
    except ImportError:
        AES = e


_ = gettext.translation("2ping", fallback=True).gettext
_pl = gettext.translation("2ping", fallback=True).ngettext

try:
    random = _pyrandom.SystemRandom()
    random_is_systemrandom = True
except AttributeError:
    random = _pyrandom
    random_is_systemrandom = False


def twoping_checksum(packet):
    """Calculate 2ping checksum on a 2ping packet

    Packet may be bytes or bytearray, return is a 32-bit int.
    Checksum is calculated as described by the 2ping protocol
    specification.
    """
    checksum = 0

    if (len(packet) % 2) == 1:
        # Convert from (possible) bytearray to bytes before appending
        packet = bytes(packet) + b"\x00"

    for i in range(0, len(packet), 2):
        checksum = checksum + (packet[i] << 8) + packet[i + 1]
        checksum = (checksum & 0xFFFF) + (checksum >> 16)

    checksum = ~checksum & 0xFFFF

    if checksum == 0:
        checksum = 0xFFFF

    return checksum


def div0(n, d):
    """Pretend we live in a world where n / 0 == 0"""
    return 0 if d == 0 else n / d


def npack(i, minimum=1):
    """Pack int to network bytes

    Takes an int and packs it to a network (big-endian) bytes.
    If minimum is specified, bytes output is padded with zero'd bytes.
    Minimum does not need to be aligned to 32-bits, 64-bits, etc.
    """
    out = bytearray()
    while i >= 256:
        out.insert(0, i & 0xFF)
        i = i >> 8
    out.insert(0, i)
    out_len = len(out)
    if out_len < minimum:
        out = bytearray(minimum - out_len) + out
    return bytes(out)


def nunpack(b):
    """Unpack network bytes to int

    Takes arbitrary length network (big-endian) bytes, and returns an
    int.
    """
    out = 0
    for x in b:
        out = (out << 8) + x
    return out


def platform_info():
    """Return a string containing platform/OS information"""
    platform_name = platform.system()
    platform_machine = platform.machine()
    platform_info = "{} {}".format(platform_name, platform_machine)
    distro_name = ""
    distro_version = ""
    distro_info = ""
    if not isinstance(distro, ImportError):
        if hasattr(distro, "name"):
            distro_name = distro.name()
        if hasattr(distro, "version"):
            distro_version = distro.version()
    if distro_name:
        distro_info = distro_name
        if distro_version:
            distro_info += " " + distro_version
        return "{} ({})".format(platform_info, distro_info)
    else:
        return platform_info


def fuzz_bytearray(data, percent, start=0, end=None):
    """Fuzz a bytearray in-place

    Each bit has a <percent> chance of being flipped.
    """
    if end is None:
        end = len(data)
    for p in range(start, end):
        xor_byte = 0
        for i in range(8):
            if random.random() < (percent / 100.0):
                xor_byte = xor_byte + (2 ** i)
        data[p] = data[p] ^ xor_byte


def fuzz_packet(packet, percent):
    """Fuzz a dumped 2ping packet

    Each bit in the opcode areas has a <percent> chance of being flipped.
    Each bit in the magic number and checksum have a <percent> / 10
    chance of being flipped.

    Returns the fuzzed packet.
    """
    packet = bytearray(packet)

    # Fuzz the entire packet
    fuzz_bytearray(packet, percent, 4)

    # Fuzz the magic number, at a lower probability
    fuzz_bytearray(packet, percent / 10.0, 0, 2)

    # Recalculate the checksum
    packet[2:4] = b"\x00\x00"
    packet[2:4] = npack(twoping_checksum(packet), 2)

    # Fuzz the recalculated checksum itself, at a lower probability
    fuzz_bytearray(packet, percent / 10.0, 2, 4)

    return bytes(packet)


def stats_time(seconds):
    """Convert seconds to ms/s/m/h/d/y time string"""

    conversion = (
        (1000, "ms"),
        (60, "s"),
        (60, "m"),
        (24, "h"),
        (365, "d"),
        (None, "y"),
    )
    out = ""
    rest = int(seconds * 1000)
    for (div, suffix) in conversion:
        if div is None:
            if out:
                out = " " + out
            out = "{}{}{}".format(rest, suffix, out)
            break
        p = rest % div
        rest = int(rest / div)
        if p > 0:
            if out:
                out = " " + out
            out = "{}{}{}".format(p, suffix, out)
        if rest == 0:
            break
    return out



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2021 Ryan Finnie
# SPDX-License-Identifier: GPL-2.0+

import argparse
import os
import socket
import sys

from . import __version__
from .utils import _, AES


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv

    if argv[0].endswith("2ping6"):
        ipv6_default = True
    else:
        ipv6_default = False

    parser = argparse.ArgumentParser(
        description="2ping ({})".format(__version__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        prog=os.path.basename(argv[0]),
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=__version__,
        help=_("report the program version"),
    )

    # Positionals
    parser.add_argument(
        "host", type=str, default=None, nargs="*", help=_("host to ping")
    )

    # ping-compatible options
    parser.add_argument(
        "--audible", "-a", dest="audible", action="store_true", help=_("audible ping")
    )
    parser.add_argument(
        "--adaptive",
        "-A",
        dest="adaptive",
        action="store_true",
        help=_("adaptive RTT ping"),
    )
    parser.add_argument(
        "--count", "-c", dest="count", type=int, help=_("number of pings to send")
    )
    parser.add_argument(
        "--flood", "-f", dest="flood", action="store_true", help=_("flood mode")
    )
    parser.add_argument(
        "--interval",
        "-i",
        dest="interval",
        type=float,
        default=1.0,
        help=_("seconds between pings"),
        metavar="SECONDS",
    )
    parser.add_argument(
        "--interface-address",
        "-I",
        dest="interface_address",
        type=str,
        action="append",
        default=[],
        help=_("interface bind address"),
        metavar="ADDRESS",
    )
    parser.add_argument(
        "--preload",
        "-l",
        dest="preload",
        type=int,
        default=1,
        help=_("number of pings to send at start"),
        metavar="COUNT",
    )
    parser.add_argument(
        "--pattern",
        "-p",
        dest="pattern",
        type=str,
        help=_("hex pattern for padding"),
        metavar="HEX_BYTES",
    )
    parser.add_argument(
        "--quiet", "-q", dest="quiet", action="store_true", help=_("quiet mode")
    )
    parser.add_argument(
        "--packetsize-compat",
        "-s",
        dest="packetsize_compat",
        type=int,
        help=_("packet size (ping compatible)"),
        metavar="BYTES",
    )
    parser.add_argument(
        "--verbose", "-v", dest="verbose", action="store_true", help=_("verbose mode")
    )
    parser.add_argument(
        "--deadline",
        "-w",
        dest="deadline",
        type=float,
        help=_("maximum run time"),
        metavar="SECONDS",
    )

    # 2ping options
    parser.add_argument("--all-interfaces", action="store_true", help=_("deprecated"))
    parser.add_argument(
        "--auth", type=str, help=_("HMAC authentication key"), metavar="KEY"
    )
    parser.add_argument(
        "--auth-digest",
        type=str,
        default="hmac-md5",
        choices=["hmac-md5", "hmac-sha1", "hmac-sha256", "hmac-crc32", "hmac-sha512"],
        help=_("HMAC authentication digest"),
        metavar="DIGEST",
    )
    parser.add_argument("--debug", action="store_true", help=_("debug mode"))
    parser.add_argument("--encrypt", type=str, help=_("Encryption key"), metavar="KEY")
    parser.add_argument(
        "--encrypt-method",
        type=str,
        default="hkdf-aes256-cbc",
        choices=["hkdf-aes256-cbc"],
        help=_("Encryption method"),
        metavar="METHOD",
    )
    parser.add_argument(
        "--fuzz", type=float, help=_("incoming fuzz percentage"), metavar="PERCENT"
    )
    parser.add_argument(
        "--inquire-wait",
        type=float,
        default=10.0,
        help=_("maximum time before loss inquiries"),
        metavar="SECONDS",
    )
    parser.add_argument("--ipv4", "-4", action="store_true", help=_("force IPv4"))
    parser.add_argument(
        "--ipv6", "-6", action="store_true", default=ipv6_default, help=_("force IPv6")
    )
    parser.add_argument("--listen", action="store_true", help=_("listen mode"))
    parser.add_argument(
        "--loopback", action="store_true", help=_("UNIX loopback test mode")
    )
    parser.add_argument(
        "--loopback-pairs",
        type=int,
        default=1,
        help=_("number of loopback pairs to create"),
        metavar="PAIRS",
    )
    parser.add_argument(
        "--max-packet-size",
        type=int,
        default=512,
        help=_("maximum packet size"),
        metavar="BYTES",
    )
    parser.add_argument(
        "--min-packet-size",
        type=int,
        default=128,
        help=_("minimum packet size"),
        metavar="BYTES",
    )
    parser.add_argument(
        "--nagios",
        type=str,
        help=_("nagios-compatible output"),
        metavar="WRTA,WLOSS%,CRTA,CLOSS%",
    )
    parser.add_argument(
        "--no-3way", action="store_true", help=_("do not send 3-way pings")
    )
    parser.add_argument(
        "--no-match-packet-size",
        action="store_true",
        help=_("do not match packet size of peer"),
    )
    parser.add_argument(
        "--no-send-version",
        action="store_true",
        help=_("do not send program version to peers"),
    )
    parser.add_argument(
        "--notice", type=str, help=_("arbitrary notice text"), metavar="TEXT"
    )
    parser.add_argument(
        "--packet-loss",
        type=str,
        help=_("percentage simulated packet loss"),
        metavar="OUT:IN",
    )
    parser.add_argument(
        "--port", type=str, default="15998", help=_("port to connect / bind to")
    )
    parser.add_argument(
        "--send-monotonic-clock",
        action="store_true",
        help=_("send monotonic clock to peers"),
    )
    parser.add_argument(
        "--send-random", type=int, help=_("send random data to peers"), metavar="BYTES"
    )
    parser.add_argument(
        "--send-time", action="store_true", help=_("send wall clock time to peers")
    )
    parser.add_argument(
        "--stats", type=float, help=_("print recurring statistics"), metavar="SECONDS"
    )
    parser.add_argument(
        "--srv", action="store_true", help=_("lookup SRV records in client mode")
    )
    parser.add_argument(
        "--srv-service",
        type=str,
        default="2ping",
        help=_("service name for SRV lookups"),
    )
    parser.add_argument(
        "--subtract-peer-host-latency",
        action="store_true",
        help=_("subtract peer host latency from RTT calculations, if sent"),
    )

    # ping-compatible ignored options
    for opt in "b|B|d|L|n|R|r|U".split("|"):
        parser.add_argument(
            "-{}".format(opt),
            action="store_true",
            dest="ignored_{}".format(opt),
            help=argparse.SUPPRESS,
        )
    for opt in "F|Q|S|t|T|M|W".split("|"):
        parser.add_argument(
            "-{}".format(opt),
            type=str,
            default=None,
            dest="ignored_{}".format(opt),
            help=argparse.SUPPRESS,
        )

    args = parser.parse_args(args=argv[1:])

    if (not args.listen) and (not args.host) and (not args.loopback):
        parser.print_help()
        parser.exit()
    if args.loopback and not hasattr(socket, "AF_UNIX"):
        parser.error("--loopback not supported on non-UNIX platforms")
    if args.nagios:
        args.quiet = True
        if (not args.count) and (not args.deadline):
            args.count = 5
        nagios_opts = args.nagios.split(",")
        if len(nagios_opts) != 4:
            parser.error(_("Invalid limits"))
        (
            args.nagios_warn_rta,
            args.nagios_warn_loss,
            args.nagios_crit_rta,
            args.nagios_crit_loss,
        ) = nagios_opts
        if args.nagios_warn_loss[-1:] != "%":
            parser.error(_("Invalid limits"))
        if args.nagios_crit_loss[-1:] != "%":
            parser.error(_("Invalid limits"))
        try:
            args.nagios_warn_loss = float(args.nagios_warn_loss[:-1])
            args.nagios_crit_loss = float(args.nagios_crit_loss[:-1])
            args.nagios_warn_rta = float(args.nagios_warn_rta)
            args.nagios_crit_rta = float(args.nagios_crit_rta)
        except ValueError as e:
            parser.error(e.message)
    if args.packetsize_compat:
        args.min_packet_size = args.packetsize_compat + 8
    if args.max_packet_size < args.min_packet_size:
        parser.error(_("Maximum packet size must be at least minimum packet size"))
    if args.max_packet_size < 64:
        parser.error(_("Maximum packet size must be at least 64"))
    args.packet_loss_in = 0
    args.packet_loss_out = 0
    if args.packet_loss:
        if ":" in args.packet_loss:
            (v_out, v_in) = args.packet_loss.split(":", 1)
            try:
                args.packet_loss_out = float(v_out)
                args.packet_loss_in = float(v_in)
            except ValueError as e:
                parser.error(e.message)
        else:
            try:
                args.packet_loss_in = args.packet_loss_out = float(args.packet_loss)
            except ValueError as e:
                parser.error(_("Packet loss: {error}").format(error=e.message))
    if args.pattern:
        if len(args.pattern) & 1:
            parser.error(_("Pattern must be full bytes"))
        if len(args.pattern) > 32:
            parser.error(_("Pattern must be 16 bytes or less"))
        args.pattern_bytes = b""
        for i in range(int(len(args.pattern) / 2)):
            a = args.pattern[(i * 2) : (i * 2 + 2)]
            try:
                b = int(a, 16)
            except ValueError as e:
                parser.error(_("Pattern: {error}").format(error=e.message))
            args.pattern_bytes += bytes([b])
    else:
        args.pattern_bytes = b"\x00"

    hmac_id_map = {
        "hmac-md5": 1,
        "hmac-sha1": 2,
        "hmac-sha256": 3,
        "hmac-crc32": 4,
        "hmac-sha512": 5,
    }
    args.auth_digest_index = hmac_id_map[args.auth_digest]

    if args.encrypt and isinstance(AES, ImportError):
        parser.error(_("Crypto module required for encryption"))

    encrypt_id_map = {"hkdf-aes256-cbc": 1}
    args.encrypt_method_index = encrypt_id_map[args.encrypt_method]

    if args.debug:
        args.verbose = True

    return args



#!/usr/bin/env python3

import os
import sys

from setuptools import setup


__version__ = "4.5.1"
assert sys.version_info > (3, 6)


def read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename), encoding="utf-8") as f:
        return f.read()


setup(
    name="2ping",
    description="2ping a bi-directional ping utility",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    version=__version__,
    license="GPLv2+",
    platforms=["Unix"],
    author="Ryan Finnie",
    author_email="ryan@finnie.org",
    url="https://www.finnie.org/software/2ping/",
    download_url="https://www.finnie.org/software/2ping/",
    packages=["twoping"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet",
        "Topic :: System :: Networking",
        "Topic :: Utilities",
    ],
    entry_points={
        "console_scripts": ["2ping = twoping.cli:main", "2ping6 = twoping.cli:main"]
    },
)



#!/usr/bin/env python3

# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2021 Ryan Finnie
# SPDX-License-Identifier: GPL-2.0+

from twoping import packets


def h(input):
    return " ".join("{:02x}".format(x) for x in input)


print("### Example 1")
print()
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x01"
print("    CLIENT: {}".format(h(packet.dump())))
print()

print("### Example 2")
print()
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xb0\x01"
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xa0\x01"
print("    SERVER: {}".format(h(packet.dump())))
print()

print("### Example 3")
print()
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xb0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xa0\x01"
print("    SERVER: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x02"
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xb0\x01"
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
print("    CLIENT: {}".format(h(packet.dump())))
print()

print("### Example 4")
print()
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x02"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x01"
)
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xb0\x02"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xa0\x02"
packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x01"
)
print("    SERVER: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x03"
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xb0\x02"
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
print("    CLIENT: {}".format(h(packet.dump())))
print()

print("### Example 5")
print()
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x02"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x01"
)
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xb0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xa0\x02"
packet.opcodes[
    packets.OpcodeInvestigationUnseen.id
] = packets.OpcodeInvestigationUnseen()
packet.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x01"
)
print("    SERVER: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x03"
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xb0\x01"
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12345
print("    CLIENT: {}".format(h(packet.dump())))
print()

print("### Example 6")
print()
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x01"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x02"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x03"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xb0\x02"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xa0\x03"
print("    SERVER: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x04"
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xb0\x02"
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 12823
print("    CLIENT: {}".format(h(packet.dump())))
print("    ... etc")
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x0a"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x01"
)
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x02"
)
print("    CLIENT: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xb0\x06"
packet.opcodes[packets.OpcodeReplyRequested.id] = packets.OpcodeReplyRequested()
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xa0\x0a"
packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x01"
)
packet.opcodes[
    packets.OpcodeInvestigationUnseen.id
] = packets.OpcodeInvestigationUnseen()
packet.opcodes[packets.OpcodeInvestigationUnseen.id].message_ids.append(
    b"\x00\x00\x00\x00\xa0\x02"
)
packet.opcodes[packets.OpcodeInvestigate.id] = packets.OpcodeInvestigate()
packet.opcodes[packets.OpcodeInvestigate.id].message_ids.append(
    b"\x00\x00\x00\x00\xb0\x02"
)
print("    SERVER: {}".format(h(packet.dump())))
packet = packets.Packet()
packet.message_id = b"\x00\x00\x00\x00\xa0\x0b"
packet.opcodes[packets.OpcodeInReplyTo.id] = packets.OpcodeInReplyTo()
packet.opcodes[packets.OpcodeInReplyTo.id].message_id = b"\x00\x00\x00\x00\xb0\x06"
packet.opcodes[packets.OpcodeRTTEnclosed.id] = packets.OpcodeRTTEnclosed()
packet.opcodes[packets.OpcodeRTTEnclosed.id].rtt_us = 13112
packet.opcodes[packets.OpcodeInvestigationSeen.id] = packets.OpcodeInvestigationSeen()
packet.opcodes[packets.OpcodeInvestigationSeen.id].message_ids.append(
    b"\x00\x00\x00\x00\xb0\x02"
)
print("    CLIENT: {}".format(h(packet.dump())))
print()



# 2ping - A bi-directional ping utility
# Copyright (C) 2010-2021 Ryan Finnie
# SPDX-License-Identifier: GPL-2.0+

import binascii
import copy
import sys

digest_size = 4


class CRC32:
    digest_size = 4
    block_size = 64
    _crc = 0

    def __init__(self, buf=None):
        if buf is not None:
            self.update(buf)

    def copy(self):
        return copy.copy(self)

    def update(self, buf):
        self._crc = binascii.crc32(buf, self._crc)

    def clear(self):
        self._crc = 0

    def digest(self):
        i = self._crc & 0xFFFFFFFF
        out = bytearray()
        while i >= 256:
            out.insert(0, i & 0xFF)
            i = i >> 8
        out.insert(0, i)
        out_len = len(out)
        if out_len < 4:
            out = bytearray(4 - out_len) + out
        return out

    def hexdigest(self):
        return self.digest().hex()


def new(buf=None):
    return CRC32(buf)


def main(argv):
    if argv is None:
        argv = sys.argv

    files = argv[1:]
    if len(files) == 0:
        files = ["-"]

    for file in files:
        c = new()
        if file == "-":
            for buf in sys.stdin.buffer.readlines():
                c.update(buf)
        else:
            with open(file, "rb") as f:
                for buf in f.readlines():
                    c.update(buf)
        print("{}\t{}".format(c.hexdigest(), file))


def module_init():
    if __name__ == "__main__":
        sys.exit(main(sys.argv))


module_init()
